import neural as nl
import padre as p
import os,shutil,glob,subprocess,datetime
from fuzzywuzzy import process

_git_ignore = [
    '*','!Data','!Data/*',
    '!Data/*/*.%s' % p.json_ext,
    '*.7z','*.dmg','*.gz','*.iso','*.jar','*.rar','*.tar','*.tgz','*.tbz','*.zip',
    '.DS_Store','.DS_Store?','._*','.Spotlight-V100','.Trashes','ehthumbs.db','Thumbs.db'
]

import imp
c = None
padre_config_file = os.path.join(p.padre_root,'padre_config.py')
if os.path.exists(padre_config_file):
    c = imp.load_source('padre_config',padre_config_file)

def commit_database(wait=True):
    '''database is stored as distributed jsons that are tracked by git -- this saves a new commit'''
    with nl.run_in(p.padre_root):
        if not os.path.exists('.git'):
            subprocess.check_call(['git','init'])
            with open('.gitignore','w') as f:
                f.write('\n'.join(_git_ignore))
        proc = subprocess.Popen(['git','add'] + glob.glob('Data/*/*.%s' % p.json_ext),stdout=subprocess.PIPE,stderr=subprocess.PIPE)
        proc.wait()
        proc = subprocess.Popen(['git','commit','-am','library commit'],stdout=subprocess.PIPE,stderr=subprocess.PIPE)
        if wait:
            proc.wait()

class commit_wrap:
    '''do a commit before and after this'''
    def __enter__(self):
        commit_database()
    
    def __exit__(self, type, value, traceback):
        commit_database()

def strip_directories(s):
    '''strip fixed leading directories and duplicate files from subjects'''
    with commit_wrap():
        new_sess = dict(s.sessions)
        for sess in new_sess:
            for label in new_sess[sess]['labels']:
                new_sess[sess]['labels'][label] = list(set([os.path.basename(x) for x in new_sess[sess]['labels'][label]]))
        s.sessions = new_sess
        s.save()

def create_subject(subject_id):
    ''' creates a new subject (loads old JSON if present and valid) '''
    with commit_wrap():
        subj = None
        if os.path.exists(p.subject_json(subject_id)):
            subj = p.Subject.load(subject_id)
            if subj==None:
                subj = p.Subject(subject_id)              
        else:
            subj = p.Subject(subject_id)
        subj.init_directories()
        subj.save()
    return subj

def delete_subject(subject_id):
    if not os.path.exists(p.trash_dir):
        os.makedirs(p.trash_dir)
    new_dir = os.path.join(p.trash_dir,'%s-%s' % (subject_id,datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S').format()))
    while os.path.exists(new_dir):
        new_dir += '_1'
    try:
        shutil.move(p.subject_dir(subject_id),new_dir)
    except OSError,IOError:
        nl.notify('Error moving subject directory %s to the trash' % subject_id,level=nl.level.error)
    try:
        del(p.subject._all_subjects[str(subject_id)])
    except KeyError:
        pass
    
class SessionExists(LookupError):
    pass

def new_session(subj,session_name):
    ''' create a new session
    
    Inserts the proper data structure into the JSON file, as well as creating
    the directory on disk.
    '''
    with commit_wrap():
        if session_name in subj._sessions:
            raise SessionExists
        session_dir = os.path.join(p.sessions_dir(subj),session_name)
        if not os.path.exists(session_dir):
            os.makedirs(session_dir)
        subj._sessions[session_name] = {'labels':{}}

def delete_session(subj,session_name,purge=False):
    ''' delete a session
    
    By default, will only delete the references to the data within the JSON file.
    If ``purge`` is given as ``True``, then it will also delete the files from
    the disk (be careful!). ``purge`` will also automatically call ``save``.'''
    with commit_wrap():
        del(subj._sessions[session_name])
        if purge:
            session_dir = os.path.join(p.sessions_dir(subj),session_name)
            if os.path.exists(session_dir):
                shutil.rmtree(session_dir)
        subj.save()

def import_raw(subj,new_dir,remove=False):
    '''copies the directory ``dir`` into the raw directory, and deletes original if ``remove`` is ``True``'''
    with commit_wrap():
        dest_dir = os.path.join(p.raw_dir(subj),os.path.split(new_dir.rstrip('/'))[1])
        if os.path.exists(dset_dir):
            raise OSError('Cannot import "%s" into subject %s - directory already exists' % (new_dir,subj))
        if remove:
            shutil.move(new_dir,dest_dir)
        else:
            shutil.copytree(new_dir,dest_dir)

def rename(subject_id,new_subject_id):
    with commit_wrap():
        subj = p.load(subject_id)
        if subj:
            try:
                os.rename(p.subject_dir(subject_id),p.subject_dir(new_subject_id))
            except OSError:
                nl.notify('Error: filesystem reported error moving %s to %s' % (subject_id,new_subject_id),level=nl.level.error)
            else:
                subj._subject_id = new_subject_id
                subj.save()
                if os.path.exists(p.subject_json(subj)):
                    try:
                        os.remove(os.path.join(p.subject_dir(subj),os.path.basename(p.subject_json(subject_id))))
                    except OSError:
                        pass
                    try:
                        del(p.subject._all_subjects[str(subject_id)])
                    except KeyError:
                        pass
    p.subject._index_one_subject(new_subject_id)

def merge_attr(f,t):
    if t==None or t=='' and not (f==None or f==''):
        t=f

dset_i = lambda label_dict,dset_fname: [x['filename'] for x in label_dict].index(dset_fname)

def merge_session(subj_from,subj_to,sess):
    if sess not in subj_to._sessions:
        subj_to._sessions[sess] = subj_from._sessions[sess]
    else:
        for k in subj_from._sessions[sess]:
            if k!= 'labels':
                if k in subj_to._sessions[sess]:
                    merge_attr(subj_from._sessions[sess][k],subj_to._sessions[sess][k])
                else:
                    subj_to._sessions[sess][k] = subj_from._sessions[sess][k]
        for label in subj_from._sessions[sess]['labels']:
            if label not in subj_to._sessions[sess]['labels']:
                subj_to._sessions[sess]['labels'][label] = []
            for dset in subj_from._sessions[sess]['labels'][label]:
                try:
                    to_i = dset_i(subj_to._sessions[sess]['labels'][label],dset['filename'])
                    subj_to._sessions[sess]['labels'][label][to_i] = dict(dset.items() + subj_to._sessions[sess]['labels'][label][to_i].items())
                except ValueError:
                    subj_to._sessions[sess]['labels'][label].append(dset)
    del(subj_from._sessions[sess])
    new_sess_dir = os.path.join(p.sessions_dir(subj_to),sess)
    from_sess_dir = os.path.join(p.sessions_dir(subj_from),sess)
    if not os.path.exists(new_sess_dir):
        os.makedirs(new_sess_dir)
    for r,ds,fs in os.walk(from_sess_dir):
        for f in fs:
            dset_f = os.path.join(new_sess_dir,f)
            if not os.path.exists(dset_f):
                os.rename(os.path.join(r,f),dset_f)
    if len(os.listdir(from_sess_dir))==0:
        os.rmdir(from_sess_dir)
    else:
        new_dir = os.path.join(p.trash_dir,'%s-%s-%s' % (subj_from,sess,datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S').format()))
        while os.path.exists(new_dir):
            new_dir += '_1'
        os.rename(from_sess_dir,new_dir)
    subj_from.save()
    subj_to.save()

def merge(subject_id_from,subject_id_into):
    nl.notify('Trying to merge %s into %s' % (subject_id_from,subject_id_into))
    with commit_wrap():
        subj_from = p.load(subject_id_from)
        subj_to = p.load(subject_id_into)
        if subj_from and subj_to:
            merge_attr(subj_from.include,subj_to.include)
            merge_attr(subj_from.notes,subj_to.notes)
            subj_to.meta = dict(subj_from.meta.items() + subj_to.meta.items())
            sess_keys = subj_from._sessions.keys()
            for sess in sess_keys:
                merge_session(subj_from,subj_to,sess)
            subj_to.save()
            delete_subject(subj_from)

def guess_label(filename,fuzzyness = 80):
    inverted_labels = {}
    for label in c.dset_labels:
        for dset in c.dset_labels[label]:
            inverted_labels[dset] = label
    label_match = process.extractOne(filename,inverted_labels.keys())
    if label_match[1] >= fuzzyness:
        label = inverted_labels[label_match[0]]
    else:
        label = 'unsorted'
    return label


def import_to_padre(subject_id,session,dsets,raw_data=[],dir_prefix=''):
    with commit_wrap():
        subj = create_subject(subject_id)
        try:
            new_session(subj,session)
        except SessionExists:
            pass
        session_dict = dict(subj._sessions[session])
        session_dict['unverified'] = True
        session_dict['date'] = datetime.datetime.strftime(nl.date_for_str(session),'%Y-%m-%d')
        for full_dset in sorted(dsets,key=lambda x:(int(os.path.basename(x).split('-')[1]),int(os.path.basename(x).split('-')[2]))):
            dset = {}
            dset['filename'] = os.path.basename(full_dset)
            if dset['filename'] not in [x.__str__(False) for x in subj.dsets(include_all=True)]:
                dset['md5'] = nl.hash(full_dset)
                dset['complete'] = True
                dset['meta'] = {}
                label = guess_label(dset['filename'].split('-')[3])
                if label not in session_dict['labels']:
                    session_dict['labels'][label] = []
                session_dict['labels'][label].append(dset)
                dset_fname = os.path.join(p.sessions_dir(subj),session,dset['filename'])
                if not os.path.exists(dset_fname):
                    shutil.move(full_dset,dset_fname)
        for raw in raw_data:
            try:
                shutil.move(os.path.join(dir_prefix,'raw',raw),p.raw_dir(subj))
            except:
                pass
        subj._sessions[session] = session_dict
        subj.save()

def dsets_identical(dset1,dset2):
    '''Tests if given datasets are identical'''
    max_tolerance = 1.0
    
    with nl.notify('Comparing %s with %s' % (dset1,dset2)):
#        info = [nl.afni.dset_info(dset) for dset in [dset1,dset2]]
#        for param in ['reps','voxel_size','voxel_dims']:
#            if getattr(info[0],param) != getattr(info[1],param):
#                nl.notify('Datasets differ in at least %s (%s vs. %s)' % (param,getattr(info[0],param),getattr(info[1],param)),level=nl.level.warning)
#                return False
        max_diff = nl.max_diff(dset1,dset2)
        if max_diff > max_tolerance:
            nl.notify('Datasets have a maximal differenence >%.1f (max_diff = %.1f)' % (max_tolerance, max_diff),level=nl.level.warning)
            return False
        
        if max_diff==0:
            nl.notify('Datasets appear to be identical')
            return True
        
        nl.notify('Datasets are minimally different (max_diff = %.1f)' % max_diff)
        return True

def sessions_identical(subj1,sess1,subj2,sess2):
    '''Tests the given sessions to make sure the datasets are the same'''
    dsets1 = [os.path.basename(str(x)) for x in subj1.dsets(session=sess1)]
    dsets2 = [os.path.basename(str(x)) for x in subj2.dsets(session=sess2)]
    dsets = list(set(dsets1+dsets2))
    return_val = True
    with nl.notify('Comparing sessions %s.%s and %s.%s:' % (subj1,sess1,subj2,sess2)):
        for dset in dsets:
            if not dsets_identical(os.path.join(p.sessions_dir(subj1),sess1,dset),os.path.join(p.sessions_dir(subj2),sess2,dset)):
                return_val = False
                continue
    return return_val

def synchronize_to_disk(subj,add_missing=True,delete_duplicates=True,delete_missing=True):
    '''Will try to clean up a subject's JSON file based on what files actually exist on disk
    
    :add_missing:           adds new JSON entries for files that exist on disk but aren't listed
    :delete_duplicates:     delete duplicate JSON entries if they refer to the same file
    :delete_missing:        delete JSON entries that have no file on disk'''
    
    def dset_in_dict(fname,l):
        return len([x for x in nl.flatten(l.values()) if fname==x['filename']])>1
    
    with nl.notify('Trying to clean up subject %s' % subj):
        s = p.load(subj)
        if s==None:
            nl.notify('Error: couldn\'t load subject %s!' % subj,level=nl.level.error)
            return False
        with commit_wrap():
            sess_on_disk = os.listdir(os.path.join(p.sessions_dir(s)))
            sess_extra_JSON = list(set(s._sessions.keys()) - set(sess_on_disk))
            sess_extra_disk = list(set(sess_on_disk) - set(s._sessions.keys()))
            if len(sess_extra_disk)>0:
                if add_missing:
                    for sess in sess_extra_disk:
                        nl.notify('Creating missing session %s' % sess,level=nl.level.warning)
                        new_session(s,sess)
                else:
                    nl.notify('Warning: found sessions on disk with no entries: %s' % (' '.join(sess_extra_disk)),level=nl.level.warning)
            if len(sess_extra_JSON)>0:
                if delete_missing:
                    for sess in sess_extra_JSON:
                        nl.notify('Removing session %s (missing from disk)' % sess, level=nl.level.warning)
                        del(s._sessions[sess])
                else:
                    nl.notify('Warning: found sessions missing from disk: %s' % (' '.join(sess_extra_disk)),level=nl.level.warning)
            for sess in s._sessions:
                with nl.notify('Checking session "%s"...' % sess):
                    new_sess = {}
                    for fname in os.listdir(os.path.join(p.sessions_dir(s),sess)):
                        if nl.is_dset(fname):
                            res = s._index_of_dset_named(fname,sess)
                            if res:
                                # At least one copy of this in the old session
                                (_,label,i) = res
                                nl.notify('Found %s' % fname)
                                if label not in new_sess:
                                    new_sess[label] = []
                                new_sess[label].append(s._sessions[sess]['labels'][label][i])
                                del(s._sessions[sess]['labels'][label][i])
                            else:
                                # File on disk, but no entry
                                if add_missing:
                                    nl.notify('Adding new entry for file %s' % fname,level=nl.level.warning)
                                    dset = {}
                                    dset['filename'] = fname
                                    full_dset = os.path.join(p.sessions_dir(s),sess,fname)
                                    dset['md5'] = nl.hash(full_dset)
                                    dset['complete'] = True
                                    dset['meta'] = {}
                                    if 'unsorted' not in new_sess:
                                        new_sess['unsorted'] = []
                                    new_sess['unsorted'].append(dset)
                    for label in s._sessions[sess]['labels']:
                        if len(s._sessions[sess]['labels'][label])>0:
                            # Leftover entries that have no file
                            for dset in s._sessions[sess]['labels'][label]:
                                if dset_in_dict(dset['filename'],new_sess):
                                    # Already have seen this dataset somewhere...
                                    if delete_duplicates:
                                        nl.notify('Deleting duplicate entry for file %s' % dset['filename'],level=nl.level.warning)
                                    else:
                                        nl.notify('Warning: found duplicate entry for file %s (leaving in place)' % dset['filename'],level=nl.level.warning)
                                        new_sess[label].append(dset)
                                else:
                                    # Entry in JSON, but no file on disk
                                    if delete_missing:
                                        nl.notify('Deleting missing dataset %s (no corresponding file on disk)' % dset['filename'],level=nl.level.warning)
                                    else:
                                        nl.notify('Warning: found entry for %s, but no corresponding file on disk (leaving empty entry in place)' % dset['filename'],level=nl.level.warning)
                                        new_sess[label].append(dset)
                    s._sessions[sess]['labels'] = new_sess
            s.save()


#!/usr/bin/env python
import padre as p
import neural as nl
import bottle
from bottle import route,view,run,post,static_file,request,redirect,error
import datetime
from dateutil.parser import parse
from fuzzywuzzy import process
import sys,os
import HTMLParser
h = HTMLParser.HTMLParser()
from padre.matching import filter_subjs

app = application = bottle.Bottle()

import imp
c = imp.load_source('padre_config',os.path.join(p.padre_root,'padre_config.py'))

web_view_root = os.path.join(sys.prefix,'padre_web')
if not os.path.exists(web_view_root):
    # Debian reports sys.prefix as '/usr', but places files in '/usr/local'
    web_view_root = os.path.join(sys.prefix,'local','padre_web')

os.chdir(web_view_root)

@app.route('/style/<filename>')
def style_file(filename):
    return static_file('views/style/%s'%filename,'.')

@app.route('/favicon.ico')
def favicon():
    return static_file('views/favicon.ico','.')

@app.route('/view_file/<subject>/<session>/<filename>')
def view_file(subject,session,filename):
    return static_file(os.path.join('Data',subject,'sessions',session,filename),root=p.padre_root)

@app.route('/')
@app.route('/index')
@view('index')
def index():
    return {'experiments':sorted(p.subject.experiments,key=lambda x: x.lower())}
    

@app.route('/list_subjects')
@app.post('/list_subjects')
@view('list_subjects')
def subjects():
    exp = request.forms.get('exp')
    if exp!='':
        subjects = p.subjects(experiment=exp)
    else:
        subjects = p.subjects()
    
    unverified = []
    for subj in subjects:
        try:
            for sess in subj._sessions:
                if 'unverified' in subj._sessions[sess]:
                    unverified.append(subj)
                    raise StopIteration
        except StopIteration:
            pass
    
    def sort_key(s):
        dates = sorted([parse(s._sessions[sess]['date']) for sess in s._sessions if 'date' in s._sessions[sess]])
        if len(dates):
            return dates[-1]
        else:
            return datetime.datetime(1,1,1)
    
#    sort_key = lambda x: sorted([parse(y[1]) for y in x[1]])[0] if len(x[1]) else datetime.datetime(1,1,1)
#    subjects = reversed(sorted([(str(s),[(sess,s.sessions[sess]['date']) for sess in s.sessions]) for s in p.subjects()],key=sort_key))
    return {
            'subjects':list(reversed(sorted(subjects,key=sort_key))),
            'unverified':list(reversed(sorted(unverified,key=sort_key))),
            'experiments':sorted(p.subject.experiments,key=lambda x: x.lower())
    }

@app.route('/edit_subject/<subject_id>')
@view('edit_subject')
def edit_subject(subject_id):
    subject = p.load(subject_id)
    unverified = []
    for sess in subject._sessions:
        if 'unverified' in subject._sessions[sess]:
            unverified.append(sess)
    sessions = list(set(subject.sessions) - set(unverified))
    return {
        'subject':subject,
        'sessions':sorted(sessions),
        'unverified':sorted(unverified),
        'experiments':sorted(p.subject.experiments,key=lambda x: x.lower())        
    }

@app.post('/save_subject')
def save_subject():
    old_subject_id = request.forms.get('old_subject_id').rstrip('/')
    new_subject_id = request.forms.get('subject_id').rstrip('/')
    if old_subject_id==new_subject_id:
        return
    if p.subject.subject_exists(new_subject_id):
        with p.maint.commit_wrap():
            p.maint.merge(old_subject_id,new_subject_id)
    else:
        with p.maint.commit_wrap():
            p.maint.rename(old_subject_id,new_subject_id)
    p.subject._index_one_subject(new_subject_id)
    if old_subject_id in p.subject._all_subjects:
        del(p.subject._all_subjects[old_subject_id])
    redirect('/edit_subject/%s' % new_subject_id)

@app.route('/edit_subject/<subject_id>/<session>')
@view('edit_session')
def edit_session(subject_id,session):
    subject = p.load(subject_id)
    return {
        'subject':subject,
        'session':session,
        'labels':sorted(p.subject.tasks,key=lambda x: x.lower()),
        'experiments':sorted(p.subject.experiments,key=lambda x: x.lower()),
        'tags': sorted(p.subject.tags,key=lambda x: x.lower())
    }

@app.post('/save_subject/<subject_id>/<session>')
@view('save_session')
def save_session(subject_id,session):
    with p.maint.commit_wrap():
        subj = p.load(subject_id)
        subj._sessions[session]['date'] = parse(request.forms.get("date")).strftime("%Y-%m-%d")
        experiment = request.forms.get("experiment")
        if experiment=='none':
            experiment = None
        if experiment=='new':
            experiment = request.forms.get("new_experiment_text")
            p.subject.experiments.add(experiment)
        subj._sessions[session]['experiment'] = experiment
        tag = request.forms.get("new_tag")
        if tag and tag!='':
            p.subject.tags.add(tag)
            if 'tags' not in subj._sessions[session]:
                subj._sessions[session]['tags'] = []
            subj._sessions[session]['tags'].append(tag)
        scan_sheet = request.files.get("scan_sheet")
        if scan_sheet != None:
            subj._sessions[session]['scan_sheet'] = scan_sheet.filename
            scan_sheet.save(os.path.join(p.sessions_dir(subj),session))
        subj._sessions[session]['notes'] = request.forms.get("notes")
        subj._sessions[session]['include'] = True if request.forms.get("include") else False
        for dset in subj.dsets(session=session,include_all=True):
            dset_fname = dset.__str__(False)
            i = [x['filename'] for x in subj._sessions[session]['labels'][dset.label]].index(dset_fname)
            subj._sessions[session]['labels'][dset.label][i]['complete'] = True if request.forms.get('complete_%s'%dset_fname) else False
            dset.complete = subj._sessions[session]['labels'][dset.label][i]['complete']
            label = request.forms.get('label_%s' % dset_fname)
            if label:
                if dset.label!=label:
                    if label=='new':
                        label = request.forms.get('label_%s_new' % dset_fname)
                        p.subject.tasks.add(label)
                        if label not in subj._sessions[session]['labels']:
                            subj._sessions[session]['labels'][label] = []
                    del(subj._sessions[session]['labels'][dset.label][i])
                    if len(subj._sessions[session]['labels'][dset.label])==0:
                        del(subj._sessions[session]['labels'][dset.label])
                    if label not in subj._sessions[session]['labels']:
                        subj._sessions[session]['labels'][label] = []
                    subj._sessions[session]['labels'][label].append(dset.__dict__())
            add_meta = request.files.get('add_meta_%s' % dset_fname)
            if add_meta:
                meta_type = request.forms.get('meta_type_%s'%dset_fname)
                subj._sessions[session]['labels'][dset.label][i]['meta'][meta_type] = add_meta.filename
                try:
                    add_meta.save(os.path.join(p.sessions_dir(subj),session))
                except IOError:
                    pass
        if 'unverified' in subj._sessions[session]:
            del(subj._sessions[session]['unverified'])
        subj.save()
        p.subject._all_subjects[str(subj)] = subj
        new_subj_id = request.forms.get("new_subject_id")
        if new_subj_id and new_subj_id!='':
            # create_subject will load the subject if it already exists...
            new_subj = p.maint.create_subject(new_subj_id)
            if new_subj:
                p.maint.merge_session(subj,new_subj,session)
                subj.save()
                new_subj.save()
                p.subject._index_one_subject(subj)
                p.subject._index_one_subject(new_subj)
                subj = new_subj
        else:
            p.subject._index_one_subject(subj)
    redirect('/edit_subject/%s/%s' % (subj,session))
    
@app.route('/delete_tag/<subject>/<session>/<tag>')
def delete_tag(subject,session,tag):
    subj = p.load(subject)
    if subj:
        if 'tags' in subj._sessions[session] and tag in subj._sessions[session]['tags']:
            del(subj._sessions[session]['tags'][subj._sessions[session]['tags'].index(tag)])
        subj.save()
    redirect('/edit_subject/%s/%s' % (subject,session))

@app.post('/search_form')
@view('list_subjects')
def search_form():
    search_string = request.forms.get('search_field')
    subjects = filter_subjs(string=search_string)
    unverified = []
    for subj in subjects:
        for sess in subj._sessions:
            if 'unverified' in subj._sessions[sess]:
                unverified.append(subj)
    subjects = list(set(subjects) - set(unverified))
    
    return {
            'subjects':subjects,
            'unverified':unverified,
            'experiments':sorted(p.subject.experiments,key=lambda x: x.lower())
    }


if __name__ == '__main__':
    import socket
    p._include_all = True
    p.subjects()
    if 'web_server' in dir(c):
        run(server=c.web_server,host=c.web_hostname,port=c.web_port)
    else:
        run(host=c.web_hostname,port=c.web_port)



#!/usr/bin/env python
''' daemon to automatically collect and organize data into padre'''

import urllib2, re, json, os, shutil, socket, tarfile, tempfile, fnmatch
from dateutil.parser import parse
import neural as nl
import padre as p
from fuzzywuzzy import process

import_location = os.path.join(p.padre_root,'Import')
processed_location = os.path.join(p.padre_root,'Processed')
import_log_file = os.path.join(import_location,'import_log.json')

import imp
c = imp.load_source('padre_config',os.path.join(p.padre_root,'padre_config.py'))

import_log = {}
if not os.path.exists(import_location):
    os.makedirs(import_location)
if os.path.exists(import_log_file):
    with open(import_log_file) as f:
        import_log = json.loads(f.read())

def save_log():
    with open(import_log_file,'w') as f:
        f.write(json.dumps(import_log))    

def import_archive(full_file,subject_guess,slice_order='alt+z',sort_order='zt'):
    tmp_dir = tempfile.mkdtemp()
    try:
        tmp_location = os.path.join(tmp_dir,'_tmp_unarchive')
        out_dir = os.path.join(processed_location,subject_guess)
        for d in [tmp_location,tmp_location + '-sorted']:
            if os.path.exists(d):
                shutil.rmtree(d)
        import_log[full_file] = {'modified':os.path.getmtime(full_file)}
        if nl.is_archive(full_file):
            with nl.notify('uncompressing files...'):
                os.makedirs(tmp_location)
                nl.unarchive(full_file,tmp_location)
        else:
            with nl.notify('copying files...'):
                shutil.copytree(full_file,tmp_location)
        with nl.notify('sorting files...'):
            nl.dicom.organize_dir(tmp_location)
        dsets_made = {}
        if os.path.exists(tmp_location + '-sorted'):
            if not os.path.exists(os.path.join(out_dir,'raw')):
                os.makedirs(os.path.join(out_dir,'raw'))
            with nl.run_in(tmp_location + '-sorted'):
                for subdir in os.listdir('.'):
                    if 'dset_ignore_masks' in dir(c):
                        ignore = False
                        for m in c.dset_ignore_masks:
                            if m in subdir:
                                ignore = True
                        if ignore:
                            continue
                    nl.notify('creating dataset from %s' % subdir)
                    import_log[full_file][subdir] = {}
                    if not nl.dicom.create_dset(subdir,slice_order,sort_order):
                        import_log[full_file][subdir]['error'] = True
                    if os.path.exists(subdir + '.nii.gz'):
                        import_log[full_file][subdir]['complete'] = True
                        shutil.move(subdir + '.nii.gz',os.path.join(out_dir,subdir+'.nii.gz'))
                        session_name = subdir.split('-')[1]
                        if session_name not in dsets_made:
                            dsets_made[session_name] = []
                        dsets_made[session_name].append(os.path.join(out_dir,subdir+'.nii.gz'))
                    if 'complete' in import_log[full_file][subdir] and import_log[full_file][subdir]['complete']:
                        if 'error' in import_log[full_file][subdir] and import_log[full_file][subdir]['error']:
                            nl.notify('created dataset %s, but Dimon returned an error' % (subdir+'.nii.gz'),nl.level.error)
                        else:
                            nl.notify('successfully created dataset %s' % (subdir+'.nii.gz'))
                    else:
                        nl.notify('failed to create dataset from directory %s' % subdir,level=nl.level.error)
            nl.notify('moving raw data...')
            raw_data = []
            for subdir in os.listdir(tmp_location + '-sorted'):
                out_file = os.path.join(out_dir,'raw',subdir + '.tgz')
                if not os.path.exists(out_file):
                    try:
                        with tarfile.open(out_file,'w:gz') as tgz:
                            tgz.add(os.path.join(tmp_location + '-sorted',subdir))
                    except tarfile.TarError:
                        nl.notify('Error creating compressed raw directory %s' % out_file,level=nl.level.error)
                    else:
                        raw_data.append(subdir + '.tgz')
            for r,ds,fs in os.walk(tmp_location):
                for f in fs:
                    out_file = os.path.join(out_dir,'raw','unsorted',r.lstrip(tmp_location).lstrip('/'),f)
                    if not os.path.exists(out_file):
                        if not os.path.exists(os.path.dirname(out_file)):
                            os.makedirs(os.path.dirname(out_file))
                        shutil.move(os.path.join(r,f),out_file)
                        raw_data.append(os.path.join('unsorted',r.lstrip(tmp_location).lstrip('/'),f))
            nl.notify('importing into padre...')
            for session in dsets_made:
                p.maint.import_to_padre(subject_guess,session,dsets_made[session],raw_data,out_dir)
        else:
            nl.notify('didn\'t find any files...')
    finally:
        shutil.rmtree(tmp_dir,True)
    save_log()
    return dsets_made



def unpack_new_archives(pi):
    new_files = []
    with nl.notify('Scanning PI %s for new archives' % pi):
        for root,dirs,files in os.walk(os.path.join(import_location,pi)):
            for fname in files + dirs:
                full_file = os.path.join(root,fname)
                if full_file not in import_log:
                    # Add in a check for the modification date
                    if any([fnmatch.fnmatch(fname,m) for m in c.archive_masks]):
                        with nl.notify('Found new archive "%s"' % full_file):
                            subject_guess = os.path.basename(os.path.dirname(full_file))
                            nl.notify('guessing the subject number is %s' % subject_guess)
                            dsets = import_archive(full_file,subject_guess)
                            new_files.append({'dir':full_file.lstrip(os.path.join(import_location,pi)).lstrip('/'),'subj':subject_guess,'fname':fname,'dsets':dsets})
    return new_files

def rsync_remote(pi):
    with nl.notify('rsync\'ing data for PI %s' % pi):
        nl.run([c.rsync,c.rsync_options,'-e','ssh -i %s' % c.server_id,'%s/%s' % (c.server_name,pi),import_location])

def email_updates(updates):
    '''Send update email to selected emails when new datsets arrives
    
    Expects the following to be set in ``padre_config.py``:
    
    Required:
    
    :demon_update_emails:       list of emails to send notifications to
    :demon_update_subject:      subject line
    :demon_update_from:         email to list as "From" address
    :demon_update_smtp_server:  SMTP server
    
    Optional:
    
    :demon_update_from_name:    name to list as "From"
    :demon_update_username      username for SMTP server (optional)
    :demon_update_password      password for SMTP server (optional)
    :demon_update_template:     method that takes the updated files as a block of text
                                and returns the full text of the email as string. 
                                Allows you to add any text you want to frame the email
    '''
    if 'demon_update_emails' not in dir(c):
        return
    
    for pi in updates:
        update_text += pi + ':\n'
        for archive in updates[pi]:
            update_text += '\t%s/%s (subject = %s)' % (updates[pi]['dir'],updates[pi]['fname'],updates[pi]['subj'])
            for sess in updates[pi]['dsets']:
                update_text += '\t\t%s:' % sess
                for dset in updates[pi]['dsets'][sess]:
                    update_text += '\t\t\t- %s' % os.path.basename(dset)
    
    try:
        update_text = c.demon_update_template(update_text)
    except:
        pass
    opts = {}
    try:
        opts['from_name'] = c.demon_update_from_name
    except:
        pass
    try:
        opts['username'] = c.demon_update_username
    except:
        pass
    try:
        opts['password'] = c.demon_update_password
    except:
        pass
    
    nl.notification.enable_email(c.demon_update_from,c.demon_update_smtp_server,**opts)
    nl.notification.email(c.demon_update_emails,update_text,c.demon_update_subject)
    


if __name__ == '__main__':
    updates = {}
    for pi in c.pis:
        rsync_remote(pi)
        updates_pi = unpack_new_archives(pi)
        if len(updates_pi)>0:
            updates[pi] = updates_pi
    if len(updates)>0:
        email_updates(updates)
                
            
